import os
import logging

from cerberus import Validator

import itertools
import riscv_config.utils as utils
from riscv_config.errors import ValidationError
from riscv_config.schemaValidator import schemaValidator
import riscv_config.constants as constants
from riscv_config.utils import yaml
from riscv_config.warl import warl_interpreter

logger = logging.getLogger(__name__)


def nosset():
    '''Function to check and set defaults for all fields which are dependent on
        the presence of 'S' extension.'''
    global inp_yaml
    if 'S' in inp_yaml['ISA']:
        return {'implemented': True}
    else:
        return {'implemented': False}


def uset():
    '''Function to set defaults based on presence of 'U' extension.'''
    global inp_yaml
    if 'U' in inp_yaml['ISA']:
        return {'implemented': True}
    else:
        return {'implemented': False}


def nuset():
    '''Function to check and set defaults for all fields which are dependent on
        the presence of 'U' extension and 'N' extension.'''
    global inp_yaml
    if 'U' in inp_yaml['ISA'] and 'N' in inp_yaml['ISA']:
        return {'implemented': True}
    else:
        return {'implemented': False}


def twset():
    '''Function to check and set value for tw field in misa.'''
    global inp_yaml
    if 'S' not in inp_yaml['ISA'] and 'U' not in inp_yaml['ISA']:
        return {'implemented': False}
    else:
        return {'implemented': False}


def delegset():
    '''Function to set "implemented" value for mideleg regisrer.'''
    # return True
    global inp_yaml
    var = True
    if 'U' not in inp_yaml['ISA']:
        var = False
    elif (('U' in inp_yaml['ISA']) and
          not ('N' in inp_yaml['ISA'] or 'S' in inp_yaml['ISA'])):
        var = False

    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True and var
    if 64 in inp_yaml['supported_xlen']:
        temp['rv64']['accessible'] = True and var
    return temp


def countset():
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 'S' in inp_yaml['ISA'] or 'U' in inp_yaml["ISA"]:
        if 32 in inp_yaml['supported_xlen']:
            temp['rv32']['accessible'] = True
        if 64 in inp_yaml['supported_xlen']:
            temp['rv64']['accessible'] = True
    return temp


def regset():
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
    if 64 in inp_yaml['supported_xlen']:
        temp['rv64']['accessible'] = True
    return temp


def counterhset():
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
    return temp


def add_def_setters(schema_yaml):
    '''Function to set the default setters for various fields in the schema'''
    regsetter = lambda doc: regset()
    schema_yaml['misa']['default_setter'] = regsetter
    schema_yaml['mstatus']['default_setter'] = regsetter
    schema_yaml['mvendorid']['default_setter'] = regsetter
    schema_yaml['mimpid']['default_setter'] = regsetter
    schema_yaml['marchid']['default_setter'] = regsetter
    schema_yaml['mhartid']['default_setter'] = regsetter
    schema_yaml['mtvec']['default_setter'] = regsetter
    schema_yaml['mip']['default_setter'] = regsetter
    schema_yaml['mie']['default_setter'] = regsetter
    schema_yaml['mscratch']['default_setter'] = regsetter
    schema_yaml['mepc']['default_setter'] = regsetter
    schema_yaml['mtval']['default_setter'] = regsetter
    schema_yaml['mcountinhibit']['default_setter'] = regsetter

    # event counters
    schema_yaml['mhpmevent3']['default_setter'] = regsetter
    schema_yaml['mhpmevent4']['default_setter'] = regsetter
    schema_yaml['mhpmevent5']['default_setter'] = regsetter
    schema_yaml['mhpmevent6']['default_setter'] = regsetter
    schema_yaml['mhpmevent7']['default_setter'] = regsetter
    schema_yaml['mhpmevent8']['default_setter'] = regsetter
    schema_yaml['mhpmevent9']['default_setter'] = regsetter
    schema_yaml['mhpmevent10']['default_setter'] = regsetter
    schema_yaml['mhpmevent11']['default_setter'] = regsetter
    schema_yaml['mhpmevent12']['default_setter'] = regsetter
    schema_yaml['mhpmevent13']['default_setter'] = regsetter
    schema_yaml['mhpmevent14']['default_setter'] = regsetter
    schema_yaml['mhpmevent15']['default_setter'] = regsetter
    schema_yaml['mhpmevent16']['default_setter'] = regsetter
    schema_yaml['mhpmevent17']['default_setter'] = regsetter
    schema_yaml['mhpmevent18']['default_setter'] = regsetter
    schema_yaml['mhpmevent19']['default_setter'] = regsetter
    schema_yaml['mhpmevent20']['default_setter'] = regsetter
    schema_yaml['mhpmevent21']['default_setter'] = regsetter
    schema_yaml['mhpmevent22']['default_setter'] = regsetter
    schema_yaml['mhpmevent23']['default_setter'] = regsetter
    schema_yaml['mhpmevent24']['default_setter'] = regsetter
    schema_yaml['mhpmevent25']['default_setter'] = regsetter
    schema_yaml['mhpmevent26']['default_setter'] = regsetter
    schema_yaml['mhpmevent27']['default_setter'] = regsetter
    schema_yaml['mhpmevent28']['default_setter'] = regsetter
    schema_yaml['mhpmevent29']['default_setter'] = regsetter
    schema_yaml['mhpmevent30']['default_setter'] = regsetter
    schema_yaml['mhpmevent31']['default_setter'] = regsetter

    schema_yaml['mhpmcounter3']['default_setter'] = regsetter
    schema_yaml['mhpmcounter4']['default_setter'] = regsetter
    schema_yaml['mhpmcounter5']['default_setter'] = regsetter
    schema_yaml['mhpmcounter6']['default_setter'] = regsetter
    schema_yaml['mhpmcounter7']['default_setter'] = regsetter
    schema_yaml['mhpmcounter8']['default_setter'] = regsetter
    schema_yaml['mhpmcounter9']['default_setter'] = regsetter
    schema_yaml['mhpmcounter10']['default_setter'] = regsetter
    schema_yaml['mhpmcounter11']['default_setter'] = regsetter
    schema_yaml['mhpmcounter12']['default_setter'] = regsetter
    schema_yaml['mhpmcounter13']['default_setter'] = regsetter
    schema_yaml['mhpmcounter14']['default_setter'] = regsetter
    schema_yaml['mhpmcounter15']['default_setter'] = regsetter
    schema_yaml['mhpmcounter16']['default_setter'] = regsetter
    schema_yaml['mhpmcounter17']['default_setter'] = regsetter
    schema_yaml['mhpmcounter18']['default_setter'] = regsetter
    schema_yaml['mhpmcounter19']['default_setter'] = regsetter
    schema_yaml['mhpmcounter20']['default_setter'] = regsetter
    schema_yaml['mhpmcounter21']['default_setter'] = regsetter
    schema_yaml['mhpmcounter22']['default_setter'] = regsetter
    schema_yaml['mhpmcounter23']['default_setter'] = regsetter
    schema_yaml['mhpmcounter24']['default_setter'] = regsetter
    schema_yaml['mhpmcounter25']['default_setter'] = regsetter
    schema_yaml['mhpmcounter26']['default_setter'] = regsetter
    schema_yaml['mhpmcounter27']['default_setter'] = regsetter
    schema_yaml['mhpmcounter28']['default_setter'] = regsetter
    schema_yaml['mhpmcounter29']['default_setter'] = regsetter
    schema_yaml['mhpmcounter30']['default_setter'] = regsetter
    schema_yaml['mhpmcounter31']['default_setter'] = regsetter

    counthsetter = lambda doc: counterhset()

    schema_yaml['mhpmcounter3h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter4h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter5h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter6h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter7h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter8h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter9h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter10h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter11h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter12h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter13h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter14h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter15h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter16h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter17h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter18h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter19h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter20h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter21h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter22h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter23h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter24h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter25h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter26h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter27h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter28h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter29h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter30h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter31h']['default_setter'] = counthsetter

    schema_yaml['mcounteren']['default_setter'] = lambda doc: countset()

    schema_yaml['mcause']['default_setter'] = regsetter
    nusetter = lambda doc: nuset()
    schema_yaml['mstatus']['schema']['rv32']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['upie'][
        'default_setter'] = nusetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['upie'][
        'default_setter'] = nusetter

    usetter = lambda doc: uset()
    schema_yaml['mstatus']['schema']['rv32']['schema']['mprv'][
        'default_setter'] = usetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['mprv'][
        'default_setter'] = usetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['uxl'][
        'default_setter'] = usetter

    schema_yaml['mip']['schema']['rv32']['schema']['ueip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv64']['schema']['ueip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv32']['schema']['utip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv64']['schema']['utip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv32']['schema']['usip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv64']['schema']['usip'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv32']['schema']['ueie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv64']['schema']['ueie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv32']['schema']['utie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv64']['schema']['utie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv32']['schema']['usie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv64']['schema']['usie'][
        'default_setter'] = nusetter

    ssetter = lambda doc: nosset()
    schema_yaml['mstatus']['schema']['rv32']['schema']['sie'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['sie'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['spie'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['spie'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['spp'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['spp'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['tvm'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['tvm'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['sxl'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['tsr'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['tsr'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['mxr'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['mxr'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['sum'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['sum'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['seip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv64']['schema']['seip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['stip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv64']['schema']['stip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['ssip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv64']['schema']['ssip'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv32']['schema']['seie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv64']['schema']['seie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv32']['schema']['stie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv64']['schema']['stie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv32']['schema']['ssie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv64']['schema']['ssie'][
        'default_setter'] = ssetter
    twsetter = lambda doc: twset()
    schema_yaml['mstatus']['schema']['rv32']['schema']['tw'][
        'default_setter'] = twsetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['tw'][
        'default_setter'] = twsetter
    delegsetter = lambda doc: delegset()
    schema_yaml['medeleg']['default_setter'] = delegsetter
    schema_yaml['mideleg']['default_setter'] = delegsetter
    return schema_yaml


def trim(foo):
    '''
        Function to trim the dictionary. Any node with implemented field set to false is trimmed of all the other nodes.

        :param foo: The dictionary to be trimmed.

        :type foo: dict

        :return: The trimmed dictionary.
    '''
    keys = foo.keys()
    if 'rv32' in keys:
        if not foo['rv32']['accessible']:
            foo['rv32'] = {'accessible': False}
    if 'rv64' in keys:
        if not foo['rv64']['accessible']:
            foo['rv64'] = {'accessible': False}
    if 'implemented' in keys:
        if not foo['implemented']:
            temp = foo
            for k in list(
                    set(foo.keys()) -
                    set(['description', 'msb', 'lsb', 'implemented', 'shadow'])
            ):
                try:
                    temp.pop(k)
                except KeyError:
                    continue
            return temp
    for key in keys:
        if isinstance(foo[key], dict):
            t = trim(foo[key])
            foo[key] = t
    return foo


def groupc(test_list):
    ''' Generator function to squash consecutive numbers for wpri bits.'''
    for x, y in itertools.groupby(enumerate(test_list), lambda a: a[1] - a[0]):
        y = list(y)
        a = y[0][1]
        b = y[-1][1]
        if a == b:
            yield [a]
        else:
            yield [a, b]


def get_fields(node, bitwidth):
    fields = list(
        set(node.keys()) -
        set(['fields', 'msb', 'lsb', 'accessible', 'shadow', 'type']))
    if not fields:
        return fields
    bits = set(range(bitwidth))
    for entry in fields:
        bits -= set(range(node[entry]['lsb'], node[entry]['msb'] + 1))
    bits = list(groupc(sorted(list(bits))))
    if not bits:
        return fields
    else:
        fields.append(bits)
        return fields


def check_reset_fill_fields(spec):
    '''The check_reset_fill_fields function fills the field node with the names of the sub-fields of the register and then checks whether the reset-value of the register is a legal value. To do so, it iterates over all the subfields and extracts the corresponding field value from the reset-value. Then it checks the legality of the value according to the given field description. If the fields is implemented i.e accessible in both 64 bit and 32 bit modes, the 64 bit mode is given preference. '''
    errors = {}
    for node in spec:
       
        if isinstance(spec[node], dict):
            if spec[node]['rv32']['accessible']:
                spec[node]['rv32']['fields'] = get_fields(
                    spec[node]['rv32'], 32)
            if spec[node]['rv64']['accessible']:
                spec[node]['rv64']['fields'] = get_fields(
                    spec[node]['rv64'], 64)
            if 'reset-val' in spec[node].keys():
                reset_val = spec[node]['reset-val']
                if spec[node]['rv64']['accessible']:
                    field_desc = spec[node]['rv64']
                    bit_len = 64
                elif spec[node]['rv32']['accessible']:
                    field_desc = spec[node]['rv32']
                    bit_len = 32
                else:
                    continue
                error = []
                if not field_desc['fields']:
                    desc = field_desc['type']
                    keys = desc.keys()
                    if 'wlrl' in keys:
                        res = False
                        for entry in desc['wlrl']:
                            if ":" in entry:
                                low, high = entry.split(":")
                                if "x" in low:
                                    low = int(low, base=16)
                                    high = int(high, base=16)
                                else:
                                    low = int(low)
                                    high = int(high)
                                if reset_val >= low and reset_val <= high:
                                    res = True
                                    break
                            else:
                                if "x" in entry:
                                    val = int(entry, base=16)
                                else:
                                    val = int(entry)
                                if val == reset_val:
                                    res = True
                                    break
                        if not res:
                            error.append(
                                "Reset value doesnt match the 'wlrl' description for the register."
                            )
                    elif 'ro_constant' in keys:
                        if reset_val not in desc['ro_constant']:
                            error.append(
                                "Reset value doesnt match the 'ro_constant' description for the register."
                            )
                    elif 'ro_variable' in keys:
                        pass
                    elif "warl" in keys:
                        warl = (warl_interpreter(desc['warl']))
                        deps = warl.dependencies()
                        dep_vals = []
                        for dep in deps:
                            reg = dep.split("::")
                            if len(reg) == 1:
                                dep_vals.append(spec[reg[0]]['reset-val'])
                            else:
                                bin_str = bin(spec[reg[0]]
                                              ['reset-val'])[2:].zfill(bit_len)
                                dep_vals.append(
                                    int(bin_str[bit_len - 1 - spec[reg[0]][
                                        'rv{}'.format(bit_len
                                                     )][reg[1]]['msb']:bit_len -
                                                spec[reg[0]]['rv{}'.format(
                                                    bit_len)][reg[1]]['lsb']],
                                        base=2))
                        if (warl.islegal(hex(reset_val)[2:], dep_vals) != True):
                            error.append(
                                "Reset value doesnt match the 'warl' description for the register."
                            )
                else:
                    bin_str = bin(reset_val)[2:].zfill(bit_len)
                    for field in field_desc['fields']:
                        if isinstance(field, list):
                            for entry in field:
                                if len(entry) == 2:
                                    test_val = int(
                                        bin_str[bit_len - 1 - entry[1]:bit_len -
                                                entry[0]],
                                        base=2)
                                    if test_val != 0:
                                        error.append("WPRI bits " +
                                                     str(entry[0]) + " to " +
                                                     str(entry[1]) +
                                                     " should be zero.")
                                elif len(entry) == 1:
                                    test_val = int(bin_str[bit_len - 1 -
                                                           entry[0]],
                                                   base=2)
                                    if test_val != 0:
                                        error.append("WPRI bit " +
                                                     str(entry[0]) +
                                                     " should be zero.")
                        else:
                            test_val = int(
                                bin_str[bit_len - 1 -
                                        field_desc[field]['msb']:bit_len -
                                        field_desc[field]['lsb']],
                                base=2)
                            if field_desc[field]['implemented']:
                                desc = field_desc[field]['type']
                                keys = desc.keys()
                                if 'wlrl' in keys:
                                    res = False
                                    for entry in desc['wlrl']:
                                        if ":" in entry:
                                            low, high = entry.split(":")
                                            if "x" in low:
                                                low = int(low, base=16)
                                                high = int(high, base=16)
                                            else:
                                                low = int(low)
                                                high = int(high)
                                            if test_val >= low and test_val <= high:
                                                res = True
                                                break
                                        else:
                                            if "x" in entry:
                                                val = int(entry, base=16)
                                            else:
                                                val = int(entry)
                                            if val == test_val:
                                                res = True
                                                break
                                    if not res:
                                        error.append(
                                            "Reset value for " + field +
                                            " doesnt match the 'wlrl' description."
                                        )
                                elif 'ro_constant' in keys:
                                    if test_val not in desc['ro_constant']:
                                        error.append(
                                            "Reset value for " + field +
                                            " doesnt match the 'ro_constant' description."
                                        )
                                elif 'ro_variable' in keys:
                                    pass
                                elif "warl" in keys:
                                    warl = (warl_interpreter(desc['warl']))
                                    deps = warl.dependencies()
                                    dep_vals = []
                                    for dep in deps:
                                        reg = dep.split("::")
                                        if len(reg) == 1:
                                            dep_vals.append(
                                                spec[reg[0]]['reset-val'])
                                        else:
                                            bin_str = bin(
                                                spec[reg[0]]['reset-val']
                                            )[2:].zfill(bit_len)
                                            dep_vals.append(
                                                int(bin_str[
                                                    bit_len - 1 -
                                                    spec[reg[0]]['rv{}'.format(
                                                        bit_len
                                                    )][reg[1]]['msb']:bit_len -
                                                    spec[reg[0]]['rv{}'.format(
                                                        bit_len
                                                    )][reg[1]]['lsb']],
                                                    base=2))
                                    if (warl.islegal(
                                            hex(test_val)[2:], dep_vals) !=
                                            True):
                                        error.append(
                                            "Reset value for " + field +
                                            " doesnt match the 'warl' description."
                                        )
                            elif test_val != 0:
                                error.append("Reset value for unimplemented " +
                                             field + " cannot be non zero.")
                if error:
                    errors[node] = error
    return spec, errors


def check_specs(isa_spec, platform_spec, work_dir, logging=False):
    '''
        Function to perform ensure that the isa and platform specifications confirm
        to their schemas. The :py:mod:`Cerberus` module is used to validate that the
        specifications confirm to their respective schemas.

        :param isa_spec: The path to the DUT isa specification yaml file.

        :param platform_spec: The path to the DUT platform specification yaml file.

        :param logging: A boolean to indicate whether log is to be printed.

        :type logging: bool

        :type isa_spec: str

        :type platform_spec: str

        :raise ValidationError: It is raised when the specifications violate the
            schema rules. It also contains the specific errors in each of the fields.

        :return: A tuple with the first entry being the absolute path to normalized isa file
            and the second being the absolute path to the platform spec file.
    '''
    global inp_yaml

    if logging:
        logger.info('Input-ISA file')

    foo = isa_spec
    schema = constants.isa_schema
    """
      Read the input-isa foo (yaml file) and validate with schema-isa for feature values
      and constraints
    """
    # Load input YAML file
    if logging:
        logger.info('Loading input file: ' + str(foo))
    inp_yaml = utils.load_yaml(foo)

    # instantiate validator
    if logging:
        logger.info('Load Schema ' + str(schema))
    schema_yaml = add_def_setters(utils.load_yaml(schema))

    #Extract xlen
    xlen = inp_yaml['supported_xlen']

    # schema_yaml = add_def_setters(schema_yaml)
    validator = schemaValidator(schema_yaml, xlen=xlen)
    validator.allow_unknown = False
    validator.purge_readonly = True
    normalized = validator.normalized(inp_yaml, schema_yaml)

    # Perform Validation
    if logging:
        logger.info('Initiating Validation')
    valid = validator.validate(normalized)

    # Print out errors
    if valid:
        if logging:
            logger.info('No Syntax errors in Input ISA Yaml. :)')
    else:
        error_list = validator.errors
        raise ValidationError("Error in " + foo + ".", error_list)
    logger.info("Initiating post processing and reset value checks.")
    normalized, errors = check_reset_fill_fields(normalized)
    if errors:
        raise ValidationError("Error in " + foo + ".", errors)
    file_name = os.path.split(foo)
    file_name_split = file_name[1].split('.')
    output_filename = os.path.join(
        work_dir, file_name_split[0] + '_checked.' + file_name_split[1])
    ifile = output_filename
    outfile = open(output_filename, 'w')
    if logging:
        logger.info('Dumping out Normalized Checked YAML: ' + output_filename)
    yaml.dump(trim(normalized), outfile)

    if logging:
        logger.info('Input-Platform file')

    foo = platform_spec
    schema = constants.platform_schema
    """
      Read the input-platform foo (yaml file) and validate with schema-platform for feature values
      and constraints
    """
    # Load input YAML file
    if logging:
        logger.info('Loading input file: ' + str(foo))
    inp_yaml = utils.load_yaml(foo)
    if inp_yaml is None:
        inp_yaml = {'mtime': {'implemented': False}}

    # instantiate validator
    if logging:
        logger.info('Load Schema ' + str(schema))
    schema_yaml = utils.load_yaml(schema)

    validator = schemaValidator(schema_yaml, xlen=xlen)
    validator.allow_unknown = False
    validator.purge_readonly = True
    normalized = validator.normalized(inp_yaml, schema_yaml)

    # Perform Validation
    if logging:
        logger.info('Initiating Validation')
    valid = validator.validate(normalized)

    # Print out errors
    if valid:
        if logging:
            logger.info('No Syntax errors in Input Platform Yaml. :)')
    else:
        error_list = validator.errors
        raise ValidationError("Error in " + foo + ".", error_list)

    file_name = os.path.split(foo)
    file_name_split = file_name[1].split('.')
    output_filename = os.path.join(
        work_dir, file_name_split[0] + '_checked.' + file_name_split[1])
    pfile = output_filename
    outfile = open(output_filename, 'w')
    if logging:
        logger.info('Dumping out Normalized Checked YAML: ' + output_filename)
    yaml.dump(trim(normalized), outfile)
    return (ifile, pfile)



from cerberus import Validator
from riscv_config.warl import warl_interpreter
import riscv_config.constants as constants
import re
import os


class schemaValidator(Validator):
    ''' Custom validator for schema having the custom rules necessary for implementation and checks.'''

    def __init__(self, *args, **kwargs):
        global rv32
        global rv64
        global extensions
        global xlen
        xlen = kwargs.get('xlen')
        if 32 in xlen:
            rv32 = True
        else:
            rv32 = False
        if 64 in xlen:
            rv64 = True
        else:
            rv64 = False
        super(schemaValidator, self).__init__(*args, **kwargs)

    def _check_with_cannot_be_false_rv64(self, field, value):
        ''' Functions ensures that the field cannot be False in rv64 mode'''
        if rv64 and not value:
            self._error(field, "This field cannot be False")

    def _check_with_cannot_be_false_rv32(self, field, value):
        ''' Functions ensures that the field cannot be False in rv32 mode'''
        if rv32 and not value:
            self._error(field, "This field cannot be False")

    def _check_with_priv_version_check(self, field, value):
        '''Function to check whether the Privileged spec version specified is valid or not.'''
        if value not in constants.priv_versions:
            self._error(
                field,
                "Invalid privilege spec version. Please select one of the following- "
                + ", ".join(constants.priv_versions))

    def _check_with_user_version_check(self, field, value):
        '''Function to check whether the User spec version specified is valid or not.'''
        if value not in constants.user_versions:
            self._error(
                field,
                "Invalid User spec version. Please select one of the following- "
                + ", ".join(constants.user_versions))

    def _check_with_capture_isa_specifics(self, field, value):
        '''
        Function to extract and store ISA specific information(such as xlen,user
        spec version and extensions present)
        and check whether the dependencies in ISA extensions are satisfied.
        '''
        global xlen
        global extensions
        extension_enc = list("00000000000000000000000000")
        if "32" in value:
            xlen = 32
            ext = value[4:]
        elif "64" in value:
            xlen = 64
            ext = value[4:]
        elif "128" in value:
            xlen = 128
            ext = value[5:]
        else:
            self._error(field, "Invalid width in ISA.")
        #ISA checks
        if any(x in value for x in "EI"):
            if 'D' in value and not 'F' in value:
                self._error(field, "D cannot exist without F.")
            if 'Q' in value and not all(x in value for x in "FD"):
                self._error(field, "Q cannot exist without F and D.")
            if 'F' in value and not "Zicsr" in value:
                self._error(field, "F cannot exist without Zicsr.")
            if 'Zam' in value and not 'A' in value:
                self._error(field, "Zam cannot exist without A.")
            if 'N' in value and not 'U' in value:
                self._error(field, "N cannot exist without U.")
            if 'S' in value and not 'U' in value:
                self._error(field, "S cannot exist without U.")
            if 'Z' in value and not self.document['User_Spec_Version'] == "2.3":
                self._error(
                    field, "Z is not supported in the User Spec given version.")
        else:
            self._error(field, "Neither of E or I extensions are present")
        #ISA encoding for future use.
        for x in "ABCDEFHIJLMNPQSTUVX":
            if (x in ext):
                extension_enc[25 - int(ord(x) - ord('A'))] = "1"
        extensions = int("".join(extension_enc), 2)
        extensions = int("".join(extension_enc), 2)

    def _check_with_rv32_check(self, field, value):
        global xlen
        if value:
            if not rv32:
                self._error(
                    field,
                    "Register cannot be implemented in rv32 mode due to unsupported xlen."
                )

    def _check_with_rv64_check(self, field, value):
        global xlen
        if value:
            if not rv64:
                self._error(
                    field,
                    "Register cannot be implemented in rv64 mode due to unsupported xlen."
                )

    def _check_with_max_length(self, field, value):
        '''Function to check whether the given value is less than the maximum value that can be stored(2^xlen-1).'''
        global xlen
        global extensions
        maxv = max(xlen)
        if value > (2**maxv) - 1:
            self._error(field, "Value exceeds max supported length")

    def _check_with_xtveccheck(self, field, value):
        '''Function to check whether the inputs in range type in mtvec are valid.'''
        global xlen
        maxv = 2**(xlen - 2)
        for list in value:
            if (len(list) > 2):
                self._error(field,
                            "Only two values are allowed in each sub list.")
            for val in list:
                if not (val < maxv):
                    self._error(field, "Invalid values.")

    def _check_with_s_check(self, field, value):
        s = 18
        if rv64 and value == True:
            mxl = format(extensions, '#066b')
            if (mxl[65 - s:66 - s] != '1'):
                self._error(field, "S is not present(64)")

        elif rv32 and value == True:
            mxl = format(extensions, '#034b')
            if (mxl[33 - s:34 - s] != '1'):
                self._error(field, "S is not present(32)")

    def _check_with_n_check(self, field, value):
        n = 13
        if rv64:
            mxl = format(extensions, '#066b')
            if (mxl[65 - n:66 - n] != '1'):
                self._error(field, "N is not present")

        elif rv32:
            mxl = format(extensions, '#034b')
            if (mxl[33 - n:34 - n] != '1'):
                self._error(field, "N is not present")

    def _check_with_mdeleg_checks(self, field, value):
        if rv32:
            if (value['rv32']['accessible'] == True and
                (not 'S' in self.document['ISA'] and
                 not 'N' in self.document['ISA'])):
                value['rv32']['accessible'] = False
                self._error(field, "S and N are not present(32)")

        if rv64:
            if (value['rv64']['accessible'] == True and
                (not 'S' in self.document['ISA'] and
                 not 'N' in self.document['ISA'])):
                value['rv64']['accessible'] = False
                self._error(field, "S and N are not present(64)")

    def _check_with_ndeleg_checks(self, field, value):
        if rv32:
            if (value['rv32']['accessible'] == True and
                    not 'N' in self.document['ISA']):
                value['rv32']['accessible'] = False
                self._error(field, "N is not present(32)")

        if rv64:
            if (value['rv64']['accessible'] == True and
                    not 'N' in self.document['ISA']):
                value['rv64']['accessible'] = False
                self._error(field, "N is not present(64)")

    def _check_with_xcause_check(self, field, value):
        '''Function to verify the inputs for mcause.'''
        if (min(value) < 16):
            self._error(
                field, "Invalid platform specific values for exception cause.")

    def _check_with_wr_illegal(self, field, value):
        pr = 0
        pri = 0
        if 'warl' not in value.keys():
            return
        for i in range(len(value['warl']['legal'])):
            if ' -> ' in value['warl']['legal'][i]:
                pr = 1
                break
        if value['warl']['wr_illegal'] != None:
            for i in range(len(value['warl']['wr_illegal'])):
                split = re.findall(r'^\s*\[(\d)]\s*',
                                   value['warl']['wr_illegal'][i])
                if split != []:
                    pri = 1
                    break
        if value['warl']['dependency_fields'] != []:
            l = (len(value['warl']['legal']))
            f = 0
            for i in range(l):
                if "bitmask" in value['warl']['legal'][i]:
                    f = 1
                    splits = re.findall(
                        r'(\[\d\])\s*->\s*.*\s*\[.*\]\s*{}\s*\[.*?[,|:].*?]'.
                        format("bitmask"), value['warl']['legal'][i])
                    if value['warl']['wr_illegal'] != None:
                        for j in range(len(value['warl']['wr_illegal'])):
                            if splits[0] in value['warl']['wr_illegal'][j]:
                                self._error(
                                    field,
                                    "illegal value does not exist for the given mode{}(bitmask)"
                                    .format(splits[0]))
            if f == 0:
                pass

        elif value['warl']['dependency_fields'] == [] and pr == 1:
            self._error(field, "no mode must exist(legal)")
        elif value['warl']['dependency_fields'] == [] and len(
                value['warl']['legal']) != 1:
            self._error(field, "There should be only one legal value")
        elif value['warl']['dependency_fields'] == [] and pri == 1:
            self._error(field, "no mode must exist(illlegal)")
        elif value['warl']['dependency_fields'] == [] and len(
                value['warl']['legal']
        ) == 1 and value['warl']['wr_illegal'] != None and "bitmask" in value[
                'warl']['legal'][0]:
            self._error(field, "illegal value cannot exist")

    def _check_with_key_check(self, field, value):
        if value['base']['type']['warl']['dependency_fields'] != []:
            par = re.split(
                "::", value['base']['type']['warl']['dependency_fields'][0])
            if not par[1] in value:
                self._error(field, " {} not present".format(par[1]))

    def _check_with_medeleg_reset(self, field, value):
        global xlen
        s = format(value, '#{}b'.format(xlen[0] + 2))
        if (s[-11:-10]) != '0' and value >= int("0x400", 16):
            self._error(field, " 11th bit must be hardwired to 0")

    def _check_with_sedeleg_reset(self, field, value):
        global xlen
        s = format(value, '#{}b'.format(xlen[0] + 2))
        if (s[-11:-8]) != '000' and value >= int("400", 16):
            self._error(field, " 11,10,9 bits should be hardwired to 0")



import re
import os


class warl_interpreter():
    global val
    global bitsum
    global dep_val

    def __init__(self, warl):
        ''' the warl_description in the yaml is given as input to the constructor '''
        self.val = warl

    def prevsum(self, i, k):
        sump = 0
        for j in range(i + 1):
            sump = sump + k[j]
        return sump

    def dependencies(self):
        dp = []
        for i in self.val['dependency_fields']:
            dp.append(i)
        self.dep_val = dp
        return dp

    def islegal(self, value, dependency_vals=[]):
        '''The function takes a range(defined as a 2 tuple list) and an optional(optional incase there are no dependencies) list containing the
        values of the corresponding dependency fields and checks whether the range of
        values is legal. Returns true if its a legal range and false otherwise.
        '''
        flag = 0
        hexa = "0x"
        xf = []
        x = []
        ch = []
        r1 = []
        r2 = []
        z = []
        l1 = []
        ch2 = []
        k = []
        cha = []
        flag1 = 0
        v = ""
        j = 0
        if self.dep_val != [] and dependency_vals != []:
            for i in range(len(self.val['legal'])):
                mode = re.findall('\[(\d)\]\s*->\s*', self.val['legal'][i])
                for i1 in range(len(dependency_vals)):
                    if dependency_vals[i1] == int(mode[i1]):
                        j = i
                        flag1 = 1
                        break
            if flag1 == 0:
                print(self.dep_val, dependency_vals)
                print("Dependency vals do not match")
                exit()
        else:
            if len(self.val['legal']) != 1:
                print("There cannot be more than one legal value")
                exit()
            else:
                j = 0
        inp1 = self.val['legal'][j]
        nodename = re.findall(r'(?:\[.+?\] -> )*\s*(.+?)\[.+?\]\s*in\s*\[.+?\]',
                              inp1)
        nodename = list(dict.fromkeys(nodename))
        if (len(nodename) == 2 and not "bitmask" in nodename.values()):
            print('Wrong Syntax: Node name is different')
            print(nodename)
            exit()
        if not "bitmask" in inp1:
            splits = re.findall(
                r'(?:{0}\[(\d.?)+:(\d.?)\])\s*in\s*\[(.+?)\]'.format(
                    nodename[0]), inp1)
            for i in range(len(splits)):
                a = int(splits[i][0])
                b = int(splits[i][1])
                ch.append(list(range(b, a + 1)))
                z.append(a - b + 1)
                if ":" in splits[i][2]:
                    y = re.split("\:", splits[i][2])
                    r1.append(int(y[0], 16))  #starting ranges
                    r2.append(int(y[1], 16))  #ending ranges
                elif "," in splits[i][2]:
                    y = re.split("\,", splits[i][2])
                    for a in y:
                        l1.append(int(a, 16))
                    r1.append(l1)
                    r2.append(l1)
                else:
                    r1.append(int(splits[i][2], 16))
                    r2.append(int(splits[i][2], 16))
            for i in range(len(splits)):
                ch2.append(int(splits[i][0]))
                ch2.append(int(splits[i][1]))

            for i in range(len(ch)):
                cha = cha + ch[i]
            for i in range(min(ch2), max(ch2) + 1):
                if i not in cha:
                    print("Bits missing error ")
                    exit()
            for i in range(len(ch)):
                for j in range(i + 1, len(ch)):
                    a_set = set(ch[i])
                    b_set = set(ch[j])
                    if (a_set & b_set):
                        print("Overlapping error")
                        exit()
            for i in z:
                if (i % 4 == 0):
                    k.append(int(i / 4))
                else:
                    k.append(int(i / 4) + 1)
            for i in range(0, len(k)):
                if ":" in splits[i][2]:
                    check = re.split(":", splits[i][2])
                    if len(check[0]) - 2 > k[i] or len(check[1]) - 2 > k[i]:
                        print("invalid range (given range exceeds)")
                        exit()
                elif isinstance(r1[i], list) == True:
                    for p in range(len(r1[i])):
                        if len(hex(r1[i][p])) - 2 > k[i]:
                            print("invalid comma seperated (exceeding)")
                            exit()
                else:
                    if len(splits[i][2]) - 2 > k[i]:
                        print("invalid fixed value (exceeding)")
                        exit()

            sum = 0
            for i in range(len(k)):
                sum = sum + k[i]

            self.bitsum = sum

            if (len(value) > sum):
                return False
            elif len(value) < sum:
                for i in range(sum - len(value)):
                    v = v + "0"
            value = v + value
            for i in range(len(k)):
                if i == 0:
                    x.append(value[0:k[i]])
                else:
                    x.append(value[self.prevsum(i - 1, k):self.prevsum(i, k)])

            for i in range(len(r1)):
                if isinstance(r1[i], list) == False:
                    if (int(x[i], 16) in range(r1[i], r2[i] + 1)):
                        flag = 1
                    else:
                        flag = 0
                        break
                else:
                    if int(x[i], 16) in r1[i]:
                        flag = 1
                    else:
                        flag = 0
                        break

        else:
            x = re.findall(
                r'\s*.*\s*\[.*\]\s*{}\s*\[(.*?,.*?)\]'.format("bitmask"), inp1)
            z = re.findall(
                r'\s*.*\s*\[(.*)\]\s*{}\s*\[.*?,.*?\]'.format("bitmask"), inp1)
            y1 = re.split("\,", x[0])
            bitmask = int(y1[0], 16)
            fixedval = int(y1[1], 16)
            currval = int(value, 16)
            legal = (currval & bitmask) | (~bitmask & fixedval)
            if (hex(currval) == hex(legal)):
                flag = 1
            else:
                flag = 0

        if flag == 1:
            return True
        else:
            return False

    def update(self, curr_val, wr_val, dependency_vals=[]):
        ''' The function takes the current value, write value and an optional list(optional incase there are no dependencies) containing the
        values of the corresponding dependency fields and models the updation of
        the register i.e if the supplied value is legal then the value is returned, else the new value of
        the register is calculated and returned.
        '''
        flag1 = 0
        flag2 = 0
        flag3 = 0
        flag4 = 0
        j = 0
        if self.dep_val != [] and dependency_vals != []:
            for i in range(len(self.val['legal'])):
                mode1 = re.findall('\[(\d)\]\s*->\s*', self.val['legal'][i])
                for i1 in range(len(dependency_vals)):
                    if dependency_vals[i1] == int(mode1[i1]):
                        j = i
                        flag1 = 1
                        break
            if flag1 == 0:
                print("Dependency vals do not match")
                exit()
        else:
            if len(self.val['legal']) != 1:
                print("There cannot be more than one legal value")
                exit()
            else:
                j = 0
        inp1 = self.val['legal'][j]
        flag1 = 0
        if self.dep_val != [] and dependency_vals != []:
            for i in range(len(self.val['legal'])):
                if "bitmask" in self.val['legal'][i]:
                    flag4 = 1
                else:
                    flag4 = 0
                    break

            if flag4 == 0 and not "bitmask" in inp1:
                for i in range(len(self.val['wr_illegal'])):
                    mode = re.findall('\[(\d)\]\s*.*->\s*',
                                      self.val['wr_illegal'][i])
                    op = re.findall(r'in\s*\[(.*?)\]',
                                    self.val['wr_illegal'][i])
                    if op != []:
                        z = re.split("\:", op[0])
                    for i1 in range(len(dependency_vals)):
                        if dependency_vals[i1] == int(mode[i1]):
                            if op != []:
                                if int(wr_val,
                                       16) in range(int(z[0], 16),
                                                    int(z[1], 16)):
                                    j = i
                                    flag1 = 1
                                    break
                            else:
                                j = i
                                flag1 = 1
                                break
                    if flag1 == 1:
                        break
                inp = self.val['wr_illegal'][j]
            else:
                inp = "no value"
                flag1 = 1

        elif self.dep_val == [] and dependency_vals == [] and not "bitmask" in self.val[
                'legal'][0]:
            for i in range(len(self.val['wr_illegal'])):
                op = re.findall(r'\s*wr_val\s*in\s*\[(.*?)\]',
                                self.val['wr_illegal'][i])
                if op != []:
                    z = re.split("\:", op[0])
                    if int(wr_val, 16) in range(int(z[0], 16), int(z[1], 16)):
                        j = i
                        flag1 = 1
                        break

                else:
                    j = i
                    flag1 = 1
                    break

            inp = self.val['wr_illegal'][j]

        elif self.dep_val != [] and dependency_vals != [] and len(
                self.val['legal']) == 1 and "bitmask" in self.val['legal'][0]:
            flag1 = 1

        if flag1 == 0 and dependency_vals != []:
            print("Dependency vals do not match")
            exit()
        if (self.islegal(curr_val, dependency_vals) == False):
            return "Current value must be legal"
        if (self.islegal(wr_val, dependency_vals)):
            return wr_val
        elif not "bitmask" in inp1:
            if "->" in inp:
                op2 = re.split(r'\->', inp)
                wr = op2[1]
            else:
                wr = inp.strip()
            if "0x" in wr:
                return wr.strip()
            elif wr.lower().strip() == "unchanged":
                return ("0x" + curr_val)

            elif wr.lower().strip() == "nearup":
                a = []
                flag2 = 0
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if len(l[i]) == 1:
                        a.append(abs(int(wr_val, 16) - int(l[i][0], 16)))
                for i in range(len(a) - 1, -1, -1):
                    if a[i] == min(a):
                        j = i
                        flag2 = 1
                        break
                if flag2 == 1:
                    return l[j][0]

            elif wr.lower().strip() == "neardown":
                a = []
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if len(l[i]) == 1:
                        a.append(abs(int(wr_val, 16) - int(l[i][0], 16)))
                for i in range(len(a)):
                    if a[i] == min(a):
                        j = i
                        flag2 = 1
                        break
                if flag2 == 1:
                    return l[j][0]

            elif wr.lower().strip() == "nextup":
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if int(l[i][0], 16) > int(wr_val, 16) and len(l[i]) == 1:
                        j = i
                        flag2 = 1
                        break
                if flag2 == 1:
                    return l[j][0]
                else:
                    return max(l)

            elif wr.lower().strip() == "nextdown":
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if int(l[i][0], 16) > int(wr_val, 16) and len(l[i]) == 1:
                        j = i
                        flag2 = 1
                        break
                if flag2 == 1 and j != 0:
                    return l[j - 1][0]
                else:
                    return min(l)

            elif wr.lower().strip() == "max":
                flag3 = 0
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if "," in l[i][0]:
                        flag3 = 1
                        j = i
                    else:
                        flag3 = 0
                if flag3 == 0:
                    return max(l)
                else:
                    y = re.split(",", l[j][0])
                    return y[1]

            elif wr.lower().strip() == "min":
                flag3 = 0
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if "," in l[i][0]:
                        flag3 = 1
                        j = i
                    else:
                        flag3 = 0
                if flag3 == 0:
                    return min(l)
                else:
                    y = re.split(",", l[j][0])
                    return y[0]

            elif wr.lower().strip() == "addr":
                wr = format(int(wr_val, 16),
                            '#0{}b'.format(4 * self.bitsum + 2))
                wr = wr[2:]
                if wr[0:1] == '0':
                    wr_final = '1' + wr[1:]
                elif wr[0:1] == '1':
                    wr_final = '0' + wr[1:]
                else:
                    print("Invalid binary bit")
                return hex(int(wr_final, 2))

            else:
                return "Invalid update mode"

        else:
            x = re.findall(
                r'\s*.*\s*\[.*\]\s*{}\s*\[(.*?,.*?)\]'.format("bitmask"), inp1)
            z = re.findall(
                r'\s*.*\s*\[(.*)\]\s*{}\s*\[.*?,.*?\]'.format("bitmask"), inp1)
            y = re.split("\,", x[0])
            bitmask = int(y[0], 16)
            fixedval = int(y[1], 16)
            currval = int(wr_val, 16)
            legal = ((currval & bitmask) | fixedval)
            return hex(legal)

    def legal(self, dependency_vals=[]):
        '''The function takes a range(defined as a 2 tuple list) and an optional(optional incase there are no dependencies) list containing the
        values of the corresponding dependency fields and returns the set of legal values as a list of two tuple lists.
        '''
        flag1 = 0
        j = 0
        if self.dep_val != [] and dependency_vals != []:
            for i in range(len(self.val['legal'])):
                mode = re.findall('\[(\d)\]\s*->\s*', self.val['legal'][i])
                for i1 in range(len(dependency_vals)):
                    if dependency_vals[i1] == int(mode[i1]):
                        j = i
                        flag1 = 1
                        break
                if flag1 == 1:
                    break
            if flag1 == 0 and dependency_vals != []:
                print("Dependency vals do not match")
                exit()
        else:
            if len(self.val['legal']) != 1:
                print("There cannot be more than one legal value")
                exit()
            else:
                j = 0
        inp = self.val['legal'][j]
        s = re.findall(r'in\s*\[(.*?)\]', inp)
        a = []
        b = []
        tup = []
        for i in range(len(s)):
            tup = []
            if ":" in s[i]:
                tup.append(s[i].replace(":", ","))
                a.append(tup)

            else:
                tup = []
                tup.append(s[i])
                a.append(tup)
        if not "bitmask" in inp:
            w = re.split(",", a[0][0])
            o = []
            for i in range(len(w)):
                e = w[i].strip().split()
                o.append(e)
            o.sort()
            return o
        else:
            return a
